class AsymException(Exception):
    pass


class MissingKeyException(AsymException):
    pass


class MissingSymmetricKeyException(MissingKeyException):
    message = "Missing Symmetric key. Set or generate one"


class MissingRsaKeyPairException(MissingKeyException):
    message = "Missing RSA key pair. Set or generate one to use RSA encryption"



import base64
import pysodium
from typing import Union


def generate_keys():
    keys = pysodium.crypto_box_keypair()
    return {'public_key': keys[0], 'private_key': keys[1]}


def hash_password(password, salt):
    if type(salt) is str:
        salt = base64.b64decode(salt)

    hashed_password = pysodium.crypto_pwhash(
        pysodium.crypto_box_SEEDBYTES,
        password,
        salt,
        pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
        pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE,
        pysodium.crypto_pwhash_ALG_DEFAULT
    )
    return hashed_password


def make_salt():
    salt = pysodium.randombytes(pysodium.crypto_pwhash_SALTBYTES)
    return salt


def encrypt(key: Union[str, bytes], message, use_base64=True):
    if type(key) is str:
        key = base64.b64decode(key)
    if type(message) is str:
        message = message.encode()
    nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
    combined_buffer = nonce + pysodium.crypto_secretbox(message, nonce, key)
    if use_base64:
        return base64.b64encode(combined_buffer).decode()
    return combined_buffer


def decrypt(
    key: Union[str, bytes], ciphertext: Union[str, bytes], make_text=True
):
    if type(key) is str:
        key = base64.b64decode(key)
    if type(ciphertext) is str:
        ciphertext = base64.b64decode(ciphertext)
    nonce = ciphertext[:pysodium.crypto_secretbox_NONCEBYTES]
    encrypted_message = ciphertext[pysodium.crypto_secretbox_NONCEBYTES:]
    decrypted_message = pysodium.crypto_secretbox_open(
        encrypted_message,
        nonce,
        key,
    )
    if make_text:
        return decrypted_message.decode()
    return decrypted_message


def export_private_key(key_pair, password=None) -> str:
    if password:
        salt = make_salt()
        hashed_password = hash_password(password, salt)
        ciphertext = encrypt(hashed_password, key_pair['private_key'], False)
        combined_buffer = salt + ciphertext
        return base64.b64encode(combined_buffer).decode()
    else:
        return base64.b64encode(key_pair['private_key']).decode()


def export_public_key(key_pair):
    return base64.b64encode(key_pair['public_key']).decode()


def import_key_pair(public_key: str, private_key: str, password=None):
    if password:
        pk = base64.b64decode(private_key)
        salt = pk[:pysodium.crypto_pwhash_SALTBYTES]
        ciphertext = pk[pysodium.crypto_pwhash_SALTBYTES:]
        hash = hash_password(password, salt)
        pk = decrypt(hash, ciphertext, False)
    else:
        pk = base64.b64decode(private_key)
    return {
        'public_key': base64.b64decode(public_key),
        'private_key': pk,
    }


def rsa_encrypt(
    their_public_key: Union[str, bytes],
    plaintext: str,
    use_base64=True,
) -> Union[str, bytes]:
    if type(their_public_key) is str:
        their_public_key = base64.b64decode(their_public_key)

    encrypted_buffer = pysodium.crypto_box_seal(
        plaintext.encode(), their_public_key)
    if use_base64:
        return base64.b64encode(encrypted_buffer).decode()
    return encrypted_buffer


def rsa_decrypt(
    my_public_key: Union[str, bytes],
    my_private_key: Union[str, bytes],
    encrypted_buffer: Union[str, bytes],
) -> Union[str, bytes]:
    if type(my_public_key) is str:
        my_public_key = base64.b64decode(my_public_key)
    if type(my_private_key) is str:
        my_private_key = base64.b64decode(my_private_key)
    if type(encrypted_buffer) is str:
        encrypted_buffer = base64.b64decode(encrypted_buffer)
    message = pysodium.crypto_box_seal_open(
        encrypted_buffer, my_public_key, my_private_key)

    return message.decode()


def generate_symmetric_key(use_base64=False) -> Union[bytes, str]:
    key = pysodium.randombytes(pysodium.crypto_secretbox_KEYBYTES)
    if use_base64:
        return base64.b64encode(key)
    return key



import unittest
import pysodium
from .crypto import (
    generate_keys, export_private_key, export_public_key, import_key_pair,
    hash_password, rsa_decrypt, rsa_encrypt
)
from .asymmetric_encryption import Asym
from .exceptions import (
    MissingRsaKeyPairException, MissingSymmetricKeyException)


class TestCrypto(unittest.TestCase):
    def test_make_and_export_RSA_key(self):
        key_pair = generate_keys()
        private_key = export_private_key(key_pair, "vacation")
        public_key = export_public_key(key_pair)
        self.assertEqual(len(private_key), 120)
        self.assertEqual(len(public_key), 44)

    def test_import_rsa_key(self):
        key_pair = generate_keys()
        private_key = export_private_key(key_pair)
        public_key = export_public_key(key_pair)
        imported_key_pair = import_key_pair(public_key, private_key)
        self.assertEqual(
            imported_key_pair['private_key'], key_pair['private_key'])
        self.assertEqual(
            imported_key_pair['public_key'], key_pair['public_key'])

    def test_import_rsa_key_from_base64(self):
        private_b64 = '2A28PjsDW/WpQqMXPLTIkL99VwnofGRqNL7wdJrzUyo='
        public_b64 = 'jim48h9bcjw85PcnaSjOAvg022kJ/Q4aTzxb8hPtGlE='
        imported_key_pair = import_key_pair(public_b64, private_b64)
        # Is the key usable?
        nonce = pysodium.randombytes(pysodium.crypto_box_NONCEBYTES)
        ciphertext = pysodium.crypto_box(
            'test',
            nonce,
            imported_key_pair['public_key'],
            imported_key_pair['private_key'],
        )

    def test_export_and_import_private_key_with_password(self):
        private_b64 = '2A28PjsDW/WpQqMXPLTIkL99VwnofGRqNL7wdJrzUyo='
        public_b64 = 'jim48h9bcjw85PcnaSjOAvg022kJ/Q4aTzxb8hPtGlE='
        password = 'hunter2'
        key_pair = import_key_pair(public_b64, private_b64)
        exported_public_key = export_public_key(key_pair)
        exported_private_key = export_private_key(key_pair, password)
        imported_key = import_key_pair(
            exported_public_key, exported_private_key, password)
        self.assertEqual(imported_key, key_pair)

    def test_hashes_password(self):
        password = 'hunter2'
        salt = bytes(
            [88, 240, 185, 66, 195, 101, 160, 138, 137, 78, 1, 2, 3, 4, 5, 6]
        )
        hash = hash_password(password, salt)
        self.assertEqual(hash[0], 49)

    def test_export_with_password(self):
        password = 'hunter2'
        key_pair = generate_keys()
        private_key = export_private_key(key_pair, password)
        self.assertIsNotNone(private_key)

    def test_rsa_encrypt_and_decrypt(self):
        key_pair = generate_keys()
        message = 'test Ơ'
        ciphertext = rsa_encrypt(key_pair['public_key'], message)
        self.assertNotEqual(ciphertext, message)
        decrypted_message = rsa_decrypt(
            key_pair['public_key'], key_pair['private_key'], ciphertext)
        self.assertEqual(decrypted_message, message)


class TestAsym(unittest.TestCase):
    asym: Asym = None

    def setUp(self):
        self.asym = Asym()

    def test_make_rsa_keys(self):
        keys = self.asym.make_rsa_keys()
        self.assertEqual(len(keys[0]), 44)
        self.assertEqual(len(keys[1]), 44)

    def test_make_rsa_keys_with_password(self):
        keys = self.asym.make_rsa_keys('123456')
        self.assertEqual(len(keys[0]), 120)
        self.assertEqual(len(keys[1]), 44)

    def test_import_rsa_key_pair(self):
        private_key_base64 = 'TCkVmbGlDxi9t+ZIYeYTgRLoHCe7kmW4AhfS8/VKqkI='
        public_key_base64 = 'wL+JHgcJs5aRYMjF8QmHcUVCWdE5ENzLVsKEf9V2UHM='
        keyPair = self.asym.set_key_pair(public_key_base64, private_key_base64)

    def test_import_rsa_key_pair_with_encrypted_private_key(self):
        private_key_base64 = 'HNP8i6zQS6Kj1b0tovnnySFCfLHWJ7ZBKfHCqperfAbgRXdlabRgRtTbb2ZyT25isGkg390tKGn+rE7emNgbkqvf9Qd6CgG37pclw0UUQy9TruasDA3/lQ=='
        public_key_base64 = 'KJ6WVf8VsH6D9842K+j0o8kGapyGSuT+MYwiw8MTdCY='
        password = '123456'

        # Try with bad password
        with self.assertRaises(Exception):
            self.asym.set_key_pair(
                public_key_base64, private_key_base64, "nope")

        key_pair = self.asym.set_key_pair(
            public_key_base64, private_key_base64, password)
        self.assertIsNotNone(key_pair)

    def test_encrypt_decrypt_using_rsa(self):
        plaintext = 'hi☢⛴'
        private_key_base64 = 'TCkVmbGlDxi9t+ZIYeYTgRLoHCe7kmW4AhfS8/VKqkI='
        public_key_base64 = 'wL+JHgcJs5aRYMjF8QmHcUVCWdE5ENzLVsKEf9V2UHM='

        self.asym.set_key_pair(public_key_base64, private_key_base64)
        ciphertext = self.asym.rsa_encrypt(
            self.asym.key_pair['public_key'], plaintext)
        self.assertNotEqual(ciphertext, plaintext)
        decrypted = self.asym.rsa_decrypt(ciphertext)
        self.assertEqual(decrypted, plaintext)

    def test_generate_symmetric_key(self):
        key = self.asym.make_symmetric_key()
        self.assertIsNotNone(key)

    def test_set_symmetric_key(self):
        key = self.asym.make_symmetric_key()
        self.asym.set_symmetric_key(key)
        self.assertIsNotNone(self.asym.symmetric_key)

    def test_get_and_set_encrypted_shared_symmetric_key(self):
        private_key_base64 = 'TCkVmbGlDxi9t+ZIYeYTgRLoHCe7kmW4AhfS8/VKqkI='
        public_key_base64 = 'wL+JHgcJs5aRYMjF8QmHcUVCWdE5ENzLVsKEf9V2UHM='

        self.asym.set_key_pair(public_key_base64, private_key_base64)
        key = self.asym.make_symmetric_key()
        encrypted_key = self.asym.get_encrypted_symmetric_key(
            self.asym.get_public_key())
        self.assertNotEqual(encrypted_key, key)
        self.asym.set_symmetric_key_from_encrypted(encrypted_key)
        self.assertEqual(self.asym.get_symmetric_key(), key)

    def test_encrypt_and_decrypt_using_symmetric_key(self):
        key = 'HMS11YymAKX5z6d6/hhdvyGtj7wiTfwzaO0O3ptSHZ4='
        message = 'hello'
        self.asym.set_symmetric_key(key)
        ciphertext = self.asym.encrypt(message)
        self.assertNotEqual(ciphertext, message)
        decrypted = self.asym.decrypt(ciphertext)
        self.assertEqual(decrypted, message)

    def test_bob_and_alice_integration_test(self):
        bob = Asym()
        alice = Asym()
        message = "hello"

        # Generate both keypairs
        bob.make_rsa_keys()
        alice.make_rsa_keys()

        # Generate one (to be shared) symmetric key
        bob.make_symmetric_key()

        # Key Exchange
        encrypted_symmetric_key = bob.get_encrypted_symmetric_key(
            alice.get_public_key())
        alice.set_symmetric_key_from_encrypted(encrypted_symmetric_key)

        # Send a message
        ciphertext = bob.encrypt(message)
        self.assertNotEqual(ciphertext, message)
        # It shold be base64 (not bytes)
        self.assertIsInstance(ciphertext, str)

        decrypted_message = alice.decrypt(ciphertext)
        self.assertEqual(decrypted_message, message)



import base64
from typing import Dict, Union, cast, List, Tuple

from .exceptions import (
    MissingSymmetricKeyException, MissingRsaKeyPairException)
from .crypto import (
    decrypt, encrypt, generate_keys, export_private_key, export_public_key,
    import_key_pair, rsa_decrypt, rsa_encrypt, generate_symmetric_key
)


KeyPair = Dict[str, bytes]
StrOrBytes = Union[str, bytes]


def to_base64(input: bytes) -> str:
    return base64.b64encode(input).decode()


def from_base64(input: str) -> bytes:
    return base64.b64decode(input)


def force_to_base64(input: StrOrBytes) -> str:
    if type(input) is bytes:
        result = to_base64(cast(bytes, input))
    else:
        result = cast(str, input)
    return result


class Asym():
    key_pair: KeyPair = None
    symmetric_key: bytes = None

    def make_rsa_keys(self, password: str = None) -> Tuple[str, str]:
        """ Makes new private key and sets for internal usage
        Return [private_key, public_key] as str. If password is set,
        the private key will be encrypted and ciphertext will be returned
        """
        self.key_pair = generate_keys()
        public_key = export_public_key(self.key_pair)
        private_key = export_private_key(self.key_pair, password)
        return (private_key, public_key)

    def set_key_pair(
        self, public_key: str, private_key: str, password: str = None
    ) -> KeyPair:
        """ Import a public and private key pair for Asym object usage (RSA
        encryption and decryption)
        """
        self.key_pair = import_key_pair(public_key, private_key, password)
        return self.key_pair

    def get_public_key(self) -> str:
        """ Get base64 version of public key """
        return to_base64(self.key_pair['public_key'])

    def get_symmetric_key(self) -> str:
        """ Return symmetric key in plain (unecrypted) base64 format """
        return to_base64(self.symmetric_key)

    def rsa_encrypt(
        self, public_key: str, plaintext: StrOrBytes, use_base64=True
    ) -> StrOrBytes:
        """ Perform RSA encryption (asymmetric)
        :param public_key: "Their" public key we want to encrypt something for
        :param plaintext: plaintext string to encrypt
        :param use_base64: Default True, return in base64 format (else bytes)
        """
        plaintext = force_to_base64(plaintext)
        return rsa_encrypt(public_key, plaintext, use_base64)

    def rsa_decrypt(self, ciphertext: str):
        """ Perform RSA Decryption (asymmetric)
        :param ciphertext: ciphertext string to decrypt
        """
        if not self.key_pair['private_key']:
            raise MissingRsaKeyPairException()
        return rsa_decrypt(
            self.key_pair['public_key'],
            self.key_pair['private_key'],
            ciphertext)

    def make_symmetric_key(self) -> str:
        """ Generate random symmetric key for internal usage and return base64
        string of it
        """
        self.symmetric_key = cast(bytes, generate_symmetric_key())
        return to_base64(self.symmetric_key)

    def set_symmetric_key(self, key: StrOrBytes):
        """ Set asym symmetric key (used for secret key encryption)
        :param key: base64 or binary key
        """
        if type(key) is str:
            self.symmetric_key = from_base64(cast(str, key))
        else:
            self.symmetric_key = cast(bytes, key)

    def get_encrypted_symmetric_key(self, public_key: str) -> str:
        """ Get encrypted version of symmetric key to share
        :param public_key: "Their" public key
        """
        key = self.rsa_encrypt(public_key, self.symmetric_key)
        return cast(str, key)

    def set_symmetric_key_from_encrypted(self, ciphertext: str):
        """ Set the symmetric (shared) key from encrypted ciphertext.
        :param ciphertext: base64 string of ciphertext version of key
        """
        decrypted_key = self.rsa_decrypt(ciphertext)
        return self.set_symmetric_key(decrypted_key)

    def encrypt(self, plaintext: str) -> str:
        if not self.symmetric_key:
            raise MissingSymmetricKeyException()
        return encrypt(self.symmetric_key, plaintext)

    def decrypt(self, ciphertext: str) -> str:
        if not self.symmetric_key:
            raise MissingSymmetricKeyException()
        return decrypt(self.symmetric_key, ciphertext)
